package com.huiju.welcome.service.impl;

import com.alibaba.fastjson.JSONObject;
import com.alibaba.fastjson.serializer.SerializerFeature;
import com.baomidou.mybatisplus.core.conditions.query.QueryWrapper;
import com.huiju.welcome.common.Action;
import com.huiju.welcome.common.WebSocketServer;
import com.huiju.welcome.mapper.SysUserMapper;
import com.huiju.welcome.mapper.TaCustomerMapMapper;
import com.huiju.welcome.mapper.TaCustomerMapper;
import com.huiju.welcome.model.SysUser;
import com.huiju.welcome.model.TaCustomer;
import com.huiju.welcome.model.TaCustomerMap;
import com.huiju.welcome.model.TaDevice;
import com.huiju.welcome.model.TaPerson;
import com.huiju.welcome.model.TaSnapshotPerson;
import com.huiju.welcome.service.*;
import com.huiju.welcome.utils.*;
import com.tencentcloudapi.common.exception.TencentCloudSDKException;
import io.swagger.models.auth.In;
import lombok.extern.slf4j.Slf4j;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Service;

import java.net.HttpURLConnection;
import java.net.URL;
import java.time.Duration;
import java.time.LocalDateTime;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

@Slf4j
@Service
public class CameraServiceImpl implements ICameraService {
    @Autowired
    private ITaDeviceService taDeviceService;

    @Autowired
    TencentFaceUtils tencentFaceUtils;

    @Autowired
    AliFaceUtils aliFaceUtils;

    @Autowired
    ISysParamService iSysParamService;

    @Autowired
    ITaPersonService taPersonService;

    @Autowired
    ITaSnapshotPersonService taSnapshotPersonService;

    @Autowired
    ITaVisitingLogService visitingLogService;

    @Autowired
    ITaCustomerService iTaCustomerService;
    
    @Autowired
    private TaCustomerMapMapper taCustomerMapMapper;
    
    @Autowired
    private TaCustomerMapper taCustomerMapper;
    
    @Autowired
    private SysUserMapper sysUserMapper;
    
    @Autowired
    MiniApp miniApp;

    @Override
    public String dispatchPerson(TaPerson taPerson, TaDevice... devices) {
        List<TaDevice> taDeviceList = null;
        if (null == devices || devices.length == 0) {
            QueryWrapper<TaDevice> queryWrapper = new QueryWrapper<>();
            queryWrapper.eq("status",StatusUtils.Normal);
            queryWrapper.eq("afr",StatusUtils.BoolTrue);
            taDeviceList = taDeviceService.list(queryWrapper);
        } else {
            taDeviceList = Arrays.asList(devices);
        }

        //分发人脸到各个相机
        for(TaDevice device: taDeviceList) {
            try {
                Map<String, Object> payload = new HashMap<>();
                payload.put("personId", taPerson.getPersonId());
                payload.put("image", taPerson.getAvatar());

                Action act = Action.create(Action.TYPE_DISPATCH_FACE, payload);

                WebSocketServer.sendInfo(act.toJSON(), "device", device.getDeviceId());
            } catch (Exception e){
                e.printStackTrace();
                return "连接设备(" + device.getDeviceId() + ")人脸下发服务 失败: "+ e.getMessage();
            }
        }
//
//        // 下发到腾讯云人脸识别
//        try {
//            tencentFaceUtils.createFace(String.valueOf(taPerson.getRealId()), taPerson.getAvatar());
//        } catch (TencentCloudSDKException e) {
//            e.printStackTrace();
//            log.error("下发人脸到腾讯云失败: " + e.getMessage());
//        }

        try {
            aliFaceUtils.createFace(String.valueOf(taPerson.getRealId()), taPerson.getAvatar());
        } catch (Exception e) {
            e.printStackTrace();
            log.error("下发人脸到阿里云失败: " + e.getMessage());
        }

        return "";
    }

    @Override
    public boolean newPersonByDevice(TaSnapshotPerson taSnapshotPerson) {
        boolean workingTime = iSysParamService.isInWorkingTime(LocalDateTime.now());
        if (!workingTime) {
            taSnapshotPersonService.save(taSnapshotPerson);
            return true;
        }

        // 人像质量最低值, 低于这个值的认为人脸抓拍失败, 取值范围 1 - 100
        Float faceMinScore = iSysParamService.getFaceMinScore();
        Integer score = taSnapshotPerson.getScore();
        if (null == score || score.intValue() < faceMinScore) {
            log.error("人脸质量过低, 放弃本次抓拍人脸 {}", taSnapshotPerson.getAvatar());
            return false;
        }

        // 人脸相似度阈值, 低于设定认为没有识别成功
        // SDK 中这个值范围是 0.001 - 1, 本系统设定范围是 1 - 100
        Float samePersonThreshold = iSysParamService.getSamePersonThreshold();
        Float similarity = taSnapshotPerson.getSimilarity();

        // 识别结果
        boolean newGuy = null == similarity || similarity < (samePersonThreshold / 100);
        TaPerson taPerson = null;
        Integer personIdOrigin = null;

        if (!newGuy) {
            if (null != taSnapshotPerson.getMatchPerson()) {
                taPerson = taPersonService.getById(taSnapshotPerson.getMatchPerson());
                if (null == taPerson) {
                    // 出现这种情况, 说明阿里云的人脸库与本地不一致。
                    newGuy = true;

                    // 保留原来人员ID
                    personIdOrigin = taSnapshotPerson.getMatchPerson();
                }
            }

            // 如果是工作人员则忽略
            if (null != taPerson && null != taPerson.getTypeId() && taPerson.getTypeId().equals(CommConstant.TYPE_INNER_PERSON)) {
                taSnapshotPerson.setMatchPerson(taPerson.getRealId());
                taSnapshotPersonService.save(taSnapshotPerson);
                return true;
            }
        }

        if (!newGuy) {
            // 抓拍人与比对人是同一个人
            taSnapshotPerson.setPersonId(taSnapshotPerson.getMatchPerson());

            LocalDateTime lastTime = taPerson.getLastVisiteDate();
            boolean sameDevice = taSnapshotPerson.getDeviceId().equals(taPerson.getLastVisiteDevice());

            if (null != lastTime && sameDevice) {
                // 降噪设定
                Float dispatchTimeLimit = iSysParamService.getDispatchTimeLimit();

                // 默认 2 分钟内连续识别的都是垃圾数据
                Duration duration = Duration.between(lastTime, taSnapshotPerson.getCreateDate());
                if (duration.getSeconds() < (60 * dispatchTimeLimit)) {
                    return true;
                }
            }
        }

        if (newGuy) {
            // 新人入库, 同时写入抓拍人ID
            if (null != personIdOrigin) {
                taSnapshotPerson.setMatchPerson(personIdOrigin);
            }

            taPerson = taPersonService.newPersonByDevice(taSnapshotPerson);

            // 同步相机识别库
            dispatchPerson(taPerson);
        }

        //二号岗设备推送
        TaDevice taDevice = taDeviceService.getById(taSnapshotPerson.getDeviceId());
        if(taDevice.getAfr().equals(CommConstant.DEVICE_AFR_UNFACE)){
            sendMessageToConsult(taPerson);
    
            // 预写到访日志
            visitingLogService.newLogByDevice(taSnapshotPerson, taPerson);
    
            // 判断是否客户来访
            try {
                iTaCustomerService.judgeNewCust(taPerson);
            } catch (Exception e) {
                log.error(e.getMessage());
            }
        }else{
            // 推送消息
            taPersonService.send2Front(taPerson);
        }

        // 保存抓拍日志
        taSnapshotPersonService.save(taSnapshotPerson);

        return true;
    }
    
    /**
     * 二号岗亭推送消息给置业经理，如果是老客户还需要推送给这个客户的置业顾问。同时都要推送给吧台。
     * @param taPerson
     */
    private void sendMessageToConsult(TaPerson taPerson) {
        TaCustomerMap taCustomerMap = taCustomerMapMapper.getCustomerOf(taPerson.getPersonId());
        if (null != taCustomerMap){
            TaCustomer taCustomer = taCustomerMapper.selectById(taCustomerMap.getCustomerId());
            if (null != taCustomer){
                try {
                    //推送置业顾问
                    iTaCustomerService.notifyConsultant(taCustomer);
                    //推送置业经理
                    sendMessageToConsultManager(taCustomer);
                    //推送吧台
                    sendMessageToBar(taCustomer,null);
                } catch (Exception e) {
                    e.printStackTrace();
                }
            }
        }
        //推送人脸到二号岗
        sendMessageToMainusher(taPerson);
    }
    
    /**
     * 推送人脸数据到二号岗
     * @param taPerson
     */
    private void sendMessageToMainusher(TaPerson taPerson) {
        try {
            WebSocketServer.sendInfo(JSONObject.toJSONString(taPerson), CommConstant.WECHAT_CLIENT_MAINUSHER, null);
        } catch (Exception e) {}
    }
    
    /**
     * 推送消息去吧台
     * @param taCustomer
     */
    @Override
    public void sendMessageToBar(TaCustomer taCustomer,Integer personNum) {
        Map<String,Object> map = new HashMap<>();
        map.put("customerId",taCustomer.getCustomerId());
        map.put("customerName",taCustomer.getCustomerName());
        map.put("consultantName",taCustomer.getRealtyConsultant());
        map.put("avatar",taCustomer.getAvatar());
        map.put("visitorNum",personNum);
        map.put("visitorTime",LocalDateTime.now());
        map.put("type",CommConstant.IDENTITY_BARUSHER);
        try {
            WebSocketServer.sendInfo(JSONObject.toJSONString(map), CommConstant.WECHAT_CLIENT_MINIAPP, null);
        } catch (Exception e) {}
    }
    
    @Override
    public void sendMessageToConsultManager(TaCustomer taCustomer) {
        //获取所有的置业经理信息并推送消息
        List<SysUser> sysUserList = sysUserMapper.selectConsultManagerList(CommConstant.IDENTITY_CONSULTANT_MANAGER);
        sysUserList.stream().forEach(e -> {
            String tpl = miniApp.getTpls().getVisitor();
            String page = CommConstant.PAGE_MY_ATTRIBUTION + "?id=" + taCustomer.getCustomerId();
            String openid = e.getOpenid();
            if (null == openid) {
                return;
            }
            List<Object> data = new ArrayList<Object>(){{
                add(taCustomer.getCustomerName());
                add(taCustomer.getPhone());
                add("您有客户来访, 请及时接待!");
            }};
    
            miniApp.sendTPLMessage(tpl, openid, page, data);
        });
    }
}